from warnings import warn as _warn

from pandas import NA as _NA, DataFrame as _Df, to_datetime as _tdt

from fipiran import _api

_KNOWN_DTYPES = {
    # 'initiationDate': 'datetime64', fails on some funds
    'alpha': 'float64',
    'annualEfficiency': 'float64',
    'auditor': 'string',
    'beta': 'float64',
    'cancelNav': 'float64',
    'custodian': 'string',
    'dailyEfficiency': 'float64',
    # date cannot be set using astype, see:
    # https://github.com/pandas-dev/pandas/issues/53127
    'date': 'O',
    'dividendIntervalPeriod': 'Int64',
    'efficiency': 'float64',
    'fundSize': 'Int64',
    'fundType': 'int64',
    'guarantor': 'string',
    'guarantorSeoRegisterNo': 'Int64',
    'initiationDate': 'O',
    'insCode': 'string',
    'investedUnits': 'Int64',
    'issueNav': 'float64',
    'manager': 'string',
    'managerSeoRegisterNo': 'Int64',
    'monthlyEfficiency': 'float64',
    'name': 'string',
    'netAsset': 'Int64',
    'quarterlyEfficiency': 'float64',
    'rankLastUpdate': 'O',
    'rankOf12Month': 'float64',
    'rankOf24Month': 'float64',
    'rankOf36Month': 'float64',
    'rankOf48Month': 'float64',
    'rankOf60Month': 'float64',
    'regNo': 'string',
    'sixMonthEfficiency': 'float64',
    'smallSymbolName': 'string',
    'statisticalNav': 'float64',
    'tempGuarantorName': 'string',
    'tempManagerName': 'string',
    'typeOfInvest': 'category',
    'weeklyEfficiency': 'float64',
}


class Fund:
    __slots__ = 'reg_no'

    def __init__(self, reg_no: int | str):
        self.reg_no = reg_no

    def __repr__(self):
        return f'{type(self).__name__}({self.reg_no!r})'

    async def asset_allocation_history(self) -> _Df:
        """Return a dict where values are percentage of each kind of asset."""
        j = await _api(f'chart/portfoliochart?regno={self.reg_no}')
        df = _Df(j)
        df['date'] = _tdt(df['date'], format='ISO8601')
        return df

    async def navps_history(self, /, *, all_=True) -> _Df:
        j = await _api(
            f'chart/getfundchart?regno={self.reg_no}&showAll={str(all_).lower()}'
        )
        df = _Df(j, copy=False)
        df['date'] = _tdt(df['date'])
        df.set_index('date', inplace=True)
        return df

    async def issue_cancel_history(self, /, *, all_=True) -> _Df:
        _warn(
            '`issue_cancel_history` is deprecated, use `navps_history` instead',
            DeprecationWarning,
        )
        return await self.navps_history(all_=all_)

    async def nav_history(self, /, *, all_=True) -> _Df:
        j = await _api(
            f'chart/getfundnetassetchart?regno={self.reg_no}&showAll={str(all_).lower()}'
        )
        df = _Df(j, copy=False)
        df['date'] = _tdt(df['date'])
        df.set_index('date', inplace=True)
        return df

    async def alpha_beta(self, /, *, all_=True) -> _Df:
        j = await _api(
            f'chart/alphabetachart?regno={self.reg_no}&showAll={str(all_).lower()}'
        )
        df = _Df(j, copy=False)
        df['date'] = _tdt(df['date'])
        df.set_index('date', inplace=True)
        return df

    async def info(self) -> dict:
        return (await _api(f'fund/getfund?regno={self.reg_no}'))['item']


def _fix_website_address(df: _Df):
    # assert df['websiteAddress'].map(len).max() == 1
    df['websiteAddress'] = (
        df['websiteAddress']
        .map(lambda lst: lst[0] if lst else _NA)
        .astype('string')
    )


def _apply_types(df: _Df) -> _Df:
    col_names = df.columns
    return df.astype(
        {cn: _KNOWN_DTYPES[cn] for cn in col_names if cn in _KNOWN_DTYPES},
        copy=False,
    )


async def funds() -> _Df:
    """Return the data available at https://fund.fipiran.ir/mf/list.

    Also see fipiran.data_service.mutual_fund_list function.
    """
    j = await _api('fund/fundcompare')
    df = _apply_types(_Df(j['items'], copy=False))
    df['date'] = _tdt(df['date'], format='ISO8601')
    _fix_website_address(df)
    return df


async def fund_types() -> _Df:
    j = await _api('fund/fundtype')
    return _Df(j['items'], copy=False)


async def average_returns() -> _Df:
    """Return a Dataframe for https://fund.fipiran.ir/mf/efficiency."""
    j = await _api('fund/averagereturns')
    df = _Df(j, copy=False)
    return df.astype({'netAsset': 'Int64'}, copy=False)


async def map_data() -> _Df:
    j = await _api('fund/treemap')
    df = _apply_types(_Df(j['items'], copy=False))
    df['date'] = _tdt(df['date'], format='ISO8601')
    _fix_website_address(df)
    return df


async def dependency_graph_data() -> _Df:
    j = await _api('fund/dependencygraph')
    df = _apply_types(_Df(j['items'], copy=False))
    df['date'] = _tdt(df['date'], format='ISO8601')
    return df



from io import StringIO as _StringIO
from typing import Literal as _Literal

from jdatetime import datetime as _jdt
from lxml.html import HtmlElement as _HtmlElement, fromstring as _parse_html
from pandas import DataFrame as _Df, read_html as _rh, to_datetime as _tdt

from . import _fipiran

_KY = ''.maketrans('کی', 'كي')


class Symbol:
    __slots__ = '_inscode', 'l18', 'l30', '_symbolpara'

    def __init__(self, l18: str, inscode: int | str = None):
        self.l18 = l18
        self._symbolpara = self.l18.translate(_KY)
        self._inscode = inscode

    def __repr__(self):
        return f'{type(self).__name__}({self.l18!r})'

    def __eq__(self, other):
        return isinstance(other, Symbol) and other.l18 == self.l18

    @property
    async def inscode(self):
        if (inscode := getattr(self, '_inscode', None)) is not None:
            return inscode
        text = await _fipiran(f'Symbol?symbolpara={self._symbolpara}')
        start = text.find("'inscode': '") + 12
        end = text.find("'", start)
        self._inscode = inscode = int(text[start:end])
        return inscode

    async def price_data(self) -> dict:
        # note: some fields like Trailing P/E and Forward P/E are *currently*
        # not computed by fipiran and are always empty.
        text = await _fipiran(
            f'Symbol/_priceData?inscode={await self.inscode}'
        )
        tree: _HtmlElement = _parse_html(text)
        d = {}

        def int_or_float(s: str) -> int | float:
            try:
                return int(s)
            except ValueError:
                return float(s)

        def num(s: str) -> int | float:
            s = s.replace(',', '').strip(' ')
            if s[0] == '(':  # negative value
                return -float(s.strip('()'))
            return int_or_float(s)

        for k in (  # numerical values
            'PriceMin',
            'PriceMax',
            'PDrCotVal',
            'PriceFirst',
            'PClosing',
            'changepdr',
            'changepc',
            'prevPrice',
            'ZTotTran',
            'QTotTran5J',
            'QTotCap',
        ):
            d[k] = num(tree.xpath(f"//*[@id='{k}']/text()")[0])

        d['Deven'] = _jdt.strptime(
            tree.xpath("//td[@id='Deven']/text()")[0], '%Y/%m/%d-%H:%M:%S'
        )

        tmin, tmax = (
            tree.xpath(
                "//td[text()='قیمت مجاز']/following-sibling::td[1]/text()"
            )[0]
            .strip()
            .split(' - ')
        )
        d['tmin'] = num(tmin)
        d['tmax'] = num(tmax)

        return d

    async def best_limit_data(self) -> list[_Df]:
        text = await _fipiran(
            f'Symbol/_BestLimitData?inscode={await self.inscode}'
        )
        return _rh(_StringIO(text))

    async def reference_data(self) -> dict:
        text = await _fipiran(
            f'Symbol/_RefrenceData?symbolpara={self._symbolpara}'
        )
        tree = _parse_html(text)
        h4s = [i.strip(': ') for i in tree.xpath('.//h4/text()')]
        spans = tree.xpath('.//span')
        d = dict(zip(h4s, [i.text for i in spans]))
        # Add ISIN e.g. 'کد معاملاتی نماد', 'IRO1MSMI0001'
        k, v = spans[0].get('title').split(' :')
        d[k] = v
        return d

    async def statistic(self, days: _Literal[365, 180, 90, 30, 7]) -> _Df:
        return _rh(
            _StringIO(
                await _fipiran(
                    f'Symbol/statistic{days}?inscode={await self.inscode}'
                )
            )
        )[0]

    async def company_info(self):
        text = await _fipiran(
            f'Symbol/CompanyInfoIndex?symbolpara={self._symbolpara}'
        )
        tree = _parse_html(text)
        d = {}
        for body in tree.xpath("//*[contains(@class, 'media-body')]"):
            key = body.xpath('normalize-space(h4)').strip(': ')
            value = body.xpath('normalize-space(.//span)')
            d[key] = value
        return d

    async def price_history(self, rows: int = 365, page: int = 1) -> dict:
        j = await _fipiran(
            'Symbol/HistoryPricePaging?'
            f'symbolpara={self._symbolpara}&rows={rows}&page={page}',
            json_resp=True,
        )
        df = j['data'] = _Df(j['data'], copy=False)
        df['gDate'] = _tdt(df['gDate'])
        df.set_index('gDate', inplace=True)
        return j


async def search(term) -> list[Symbol]:
    """Note: the returned symbol objects will have `l30` attribute."""
    symbols = []
    append = symbols.append
    for result in await _fipiran('Home/AutoComplete', (('term', term),), True):
        symbol = Symbol(result['LVal18AFC'])
        symbol.l30 = result['LSoc30']
        append(symbol)
    return symbols



from functools import partial as _partial
from re import compile as _rc

from aiohutils.pd import html_to_df as _hd
from pandas import DataFrame as _Df

from . import _fipiran

_parenthesis_to_negative = _partial(_rc(r'\((\d+)\)').sub, r'-\1')


async def financial_ratios() -> _Df:
    text = await _fipiran('Codal/Ratio')
    return _hd(text)


async def profit_growth() -> _Df:
    text = await _fipiran('Codal/RoshdPos')
    return _hd(_parenthesis_to_negative(text))


async def profit_decline() -> _Df:
    text = await _fipiran('Codal/RoshdNeg')
    return _hd(_parenthesis_to_negative(text))



__version__ = '0.23.0'

from json import JSONDecodeError as _JSONDecodeError, loads as _jl
from logging import error as _error

from aiohutils.session import SessionManager
from pandas import options as _o

_o.mode.copy_on_write = True
_o.future.infer_string = True

_FIPIRAN = 'https://www.fipiran.ir/'
_YK = ''.maketrans('يك', 'یک')
_API = 'https://fund.fipiran.ir/api/v1/'


session_manager = SessionManager(
    headers={
        'Referer': 'https://fund.fipiran.ir',
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/116.0',
    },
)


async def _read(url, **kwargs) -> bytes:
    r = await session_manager.get(url, **kwargs)
    return await r.read()


async def _api(path) -> dict | list:
    r = await _read(_API + path)
    try:
        return _jl(r)
    except _JSONDecodeError:
        _error(f'{r = }')
        raise


async def _fipiran(
    path: str, params=None, json_resp=False
) -> str | dict | list:
    text = (
        (await _read(f'{_FIPIRAN}{path}', params=params))
        .decode()
        .translate(_YK)
    )
    if json_resp is True:
        return _jl(text)
    return text
